/* -*-	Mode:C++; c-basic-offset:4; tab-width:4; indent-tabs-mode:t -*- */

/*
 * Copyright (c) 2014, Damien Magoni.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Damien Magoni.
 * 4. The name of Damien Magoni may not be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DAMIEN MAGONI ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL DAMIEN MAGONI OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

// DMP for Partial Data Aggregation

#ifndef ns_dmp_h
#define ns_dmp_h

#include <cstdio>
#include <cmath>
#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <map>
#include <stack>
#include "common/agent.h"
#include "common/packet.h"
#include "common/ip.h"
#include "common/timer-handler.h"
#include "mac/mac.h"

/*
 	typedef int32_t nsaddr_t;
	typedef int32_t nsmask_t;
	struct ns_addr_t {
		int32_t addr_;
		int32_t port_;
	};
*/

const int DMP_BEACON = 1;
const int DMP_BEACON_REPLY = 2;
const int DMP_DATA = 3;

const int SENSOR_CM = 1;
const int SENSOR_CH = 2;

typedef int aggregation_type;
const aggregation_type aggregation_none = 1;
const aggregation_type aggregation_partial = 2;
const aggregation_type aggregation_random = 3;
const aggregation_type aggregation_full = 4;

struct hdr_dmp
{
	int dmp_subtype_; // not the same as ptype in common header!
	int src_address_;
	int dst_address_;
	double timestamp_;
	double x_; // sender position for spatial aggreg
	double y_;
	double speed_;
	double heading_;
	int comm_nb_;

	int aggregation_; // 0 = no, 1 = yes
	int cluster_id_;
	int sensor_type_; // identify if the sender sensor is CH or CM
	int content_type_;
	double content_value_;
	int priority_level_;

/*	int & src_address() { return (src_address_); }
	int & dst_address() { return (dst_address_); }
	double & timestamp() { return (timestamp_); }
	double & x() { return (x_); }
	double & y() { return (y_); }
	double & speed() { return (speed_); }
	double & heading() { return (heading_); }
	int & comm_nb() { return (comm_nb_); }*/

	static int offset_;
	inline static int& offset() { return offset_; }
	inline static hdr_dmp * access(const Packet * p)
	{
		return (hdr_dmp *) p->access(offset_);
	}
};

class entry
{
public:
	int priority_level;
	int content_type;
	double x, y;
	bool partial;
	std::map<double, double> values; // time, value pairs received during a period
};

class DmpAgent;

class DMPDataTimer : public TimerHandler
{
public:
	DMPDataTimer(DmpAgent * a) : TimerHandler() { a_ = a; }
	virtual void expire(Event * e);
protected:
	DmpAgent * a_;
};

class DMPLocalAggTimer : public TimerHandler
{
public:
	DMPLocalAggTimer(DmpAgent * a) : TimerHandler() { a_ = a; }
	virtual void expire(Event * e);
protected:
	DmpAgent * a_;
};

class DMPGlobalAggTimer : public TimerHandler
{
public:
	DMPGlobalAggTimer(DmpAgent * a) : TimerHandler() { a_ = a; }
	virtual void expire(Event * e);
protected:
	DmpAgent * a_;
};

class DMPBeaconTimer : public TimerHandler
{
public:
	DMPBeaconTimer(DmpAgent * a) : TimerHandler() { a_ = a; }
	virtual void expire(Event * e);
protected:
	DmpAgent * a_;
};

class demigod;
class Node;
class Mac;

class DmpAgent : public Agent
{
public:
	DmpAgent();
	int command(int argc, const char * const * argv);
	void start();
	void stat();
	virtual void recv(Packet *, Handler *);
	int hook(Packet * p); // from Mac (or LL) up to Agent
	void inject(Packet * p); // from Agent down to LL (or Mac)
	void data_timeout();
	void local_aggr_timeout();
	void global_aggr_timeout();
	void aggregate_data(Packet * p);
	void process_data(Packet * p);
	void store_data(bool local, Packet * p);
	void evaluate_partial_aggr(bool local, Packet * p);

	void init_beacon_timer();
	void beacon_timeout();
	void broadcast_beacon();
	void beacon(int address);
	void process_beacon(Packet * p);
	void process_reply(Packet * p);
	void set_x();
	void set_y();
	void set_range();
	void set_power();

    Node * node_pt;
    int node_id_;
	int node_address_; // should be the same as node_id in standard addressing case
	Mac * mac_pt;
	int mac_address_; // should be the same as node_address
	int sensor_type_;
	int cluster_id_;
	double spatial_closeness_;
	double message_period_;
	double local_wait_period_, min_local_values, cv_local_threshold;
	double global_wait_period_, min_global_values, cv_global_threshold;
	DMPDataTimer data_timer;
	DMPLocalAggTimer local_agg_timer;
	DMPGlobalAggTimer global_agg_timer;
	int epoch; // for the data timer
	int pkt_recv; // aggreg ratio
	int pkt_sent;
	int pkt_sent_fast;
	std::vector<double> e2e_delay; // used in sink only

	std::vector<entry> local_values; // they are automatically spatially correlated, only temporal correlation needs to be done
	std::vector<entry> global_values;

	int CM_nb_;
	int net_type_;
	int nodes_nb_;
	int verbose_;
	int algorithm_;
	double seed_;
	double beacon_period_;
	int min_connect_;
	double tx_max_;
	double margin_;
	double rx_thr_;
	double range_max_;
	double x;
	double y;
	double energy_used;
	double transmission_power;
	double transmission_range;
	double transmission_power_change_time;
	DMPBeaconTimer beacon_timer;
	std::vector<int> neighbors;
	std::vector<int> maxneighbors;
	std::map<double, int> communicators;
	int cc_no;
	int prefix;
	int back;
	demigod * dg;
};

class demigod
{
public:
	typedef std::vector<double> distance_vector;
	std::vector<distance_vector> distance_matrix;
	typedef std::vector<int> connect_vector;
	std::vector<connect_vector> connect_matrix;
	std::map<int, DmpAgent *> nodes;
	int cc_nb, mcc_nb;
	int cc_cn, mcc_cn;
	int net;
	int gen_pkts_sent;
	int tot_pkts_sent;

	demigod();
	void init(DmpAgent * a);
	void update(DmpAgent * a);
	void output();
	int tarjan();
	void strongly_connected_component(DmpAgent * x, int * cpt, int * m, std::stack<DmpAgent *> * P);
};

namespace dmp
{
	const unsigned long DMP_INT32_MAX = 0xffffffffUL;
	const double pi = 3.141592653589793238463;
	const double dbl_max = 1.7976931348623157E+308;

	double dist(double x1, double y1, double x2, double y2);
	int round(double d);
	int random(int inf, int sup);
	double rzto();
	bool roll(double pc_probability);

	std::vector<double> extract_vec(std::map<double, double> & mm, bool key);
	double sample_sum(std::vector<double> & sample_data);
	double sample_min(std::vector<double> & sample_data);
	double sample_max(std::vector<double> & sample_data);
	double sample_perc(double perc, std::vector<double> & sample_data);
	double sample_mean(std::vector<double> & sample_data);
	double sample_standard_deviation(std::vector<double> & sample_data);
	double sample_standard_error_of_the_mean(std::vector<double> & sample_data);
	double sample_relative_statistical_error(double confidence_level, std::vector<double> & sample_data);

	// John C. Pezzullo at johnp71@aol.com
	double StatCom(double q, int i, int j, int b);
	double StudT(double t, int n);
	double AStudT(double p, int n);

	// Mersenne Twister
	void init_genrand(unsigned long s);
	void init_by_array(unsigned long init_key[], unsigned long key_length);
	unsigned long genrand_int32(void);
	long genrand_int31(void);
	double genrand_real1(void);
	double genrand_real2(void);
	double genrand_real3(void);
	double genrand_res53(void);
}

#endif // ns_dmp_h
